// $Id: FreeEnergyModel.h 1136 2020-07-17 17:28:27Z ge $
/// \file FreeEnergyModel.h
/// \brief contains the FreeEnergyModel class
///
/// $Revision: 1136 $
/// \author Gerald Weber <gweberbh@gmail.com>
#ifndef GBC_FREEENERGYMODEL_H
#define GBC_FREEENERGYMODEL_H "$Id: FreeEnergyModel.h 1136 2020-07-17 17:28:27Z ge $"
#include "SequenceInfo.h"
#include "SequenceDataset.h"
#include "IndividualNearestNeighbours.h"
#include "valmatrix.h"

namespace gbc
{
template<class _Model=IndividualNearestNeighbours<double> >
class FreeEnergyModel
  {
  public:
    typedef _Model                                 model_type;
    typedef typename model_type::value_type        value_type;
    typedef std::valarray<value_type>              vector_type;
    typedef typename model_type::sequenceinfo_type sequenceinfo_type;
    typedef SequenceDataset<sequenceinfo_type>     sequence_dataset_type;
    typedef std::deque<std::string>                string_deque_type;
    typedef valmatrix<value_type>                  matrix_type;

    sequence_dataset_type dataset;
    model_type            *model;
    std::string           data_set_type;
    matrix_type           A, CT;                  ///< The matrices
    string_deque_type     HSnames;                ///< List of NN names corresponding to matrix A

    FreeEnergyModel(void): model() {}

   inline void verify(std::ostream& out)
      {
      dataset.calculate_and_print_verify(out);
      }

   inline void print_tex(std::ostream& out)
     {
     dataset.print_tex(out);
     }

    /// \brief Calculates the difference between the predicted and the adjusted temperature
    inline void predict_all(void)
      {
      typename sequence_dataset_type::sequence_info_deque_type::iterator si=dataset.Raw_dataset.begin();
      for(si=dataset.Raw_dataset.begin(); si != dataset.Raw_dataset.end(); si++)
        {
        model->Sequence_info=&(*si);
        model->reset_calculation();
        model->melting_temperature();
        }
      }
      
    void  make_matrices(void) 
      {
      typename sequence_dataset_type::neighbours_ref_type::map_type::const_iterator it;
      typename sequence_dataset_type::sequence_info_deque_type::iterator si=dataset.Raw_dataset.begin();
      size_t M=dataset.Raw_dataset.size();
      size_t N=2*dataset.Reduced_neighbours_set.size();
      
      A.resize(M,N);
      HSnames.resize(N);
      
      int m, n;
      for(m=0, si=dataset.Raw_dataset.begin(); si != dataset.Raw_dataset.end(); si++, m++)
        {
        for(n=0, it=dataset.Reduced_neighbours_set.begin(); it != dataset.Reduced_neighbours_set.end(); ++it, n++)
          {
          A(m,n)=value_type(si->Reduced_neighbours_set.count(it->first));
          A(m,n+N/2)=(si->temperature.adjusted+273.15)*A(m,n);  
          }
        }
      for(n=0, it=dataset.Reduced_neighbours_set.begin(); it != dataset.Reduced_neighbours_set.end(); ++it, n++)
        {
        HSnames[n]=std::string(it->first) + std::string(":enthalpy");
        HSnames[n+N/2]=std::string(it->first) + std::string(":entropy");
        }
      }

      
    //Rules are read from parameter files, make sure that this is set before read_xml
    //otherwise the rules are not applied to the dataset
    void read_rules_and_parameters_from_list_of_files(string_deque_type &par_file_deque)
      {
      dataset.read_rules_from_list_of_files(par_file_deque);
      model->get_parameters_from_list_of_files(par_file_deque);
      }


  };

}
#endif
