// $Id: GibbsMinimization.h 1136 2020-07-17 17:28:27Z ge $
/// \file GibbsMinimization.h
/// \brief contains the GibbsMinimization function class
///
/// This is a plugin class for the ParameterSearch function
/// $Revision: 1136 $
/// \author Gerald Weber <gweberbh@gmail.com>
#ifndef GBC_GIBBSMINIMIZATION_H
#define GBC_GIBBSMINIMIZATION_H "$Id: GibbsMinimization.h 1136 2020-07-17 17:28:27Z ge $"
#include "FreeEnergyModel.h"

namespace gbc
{
template<class _NNModel=IndividualNearestNeighbours<double> >
/// \brief This class requires an xml file of data to work
class GibbsMinimization
  {
  public:
    typedef _NNModel                                  model_type;
    typedef typename model_type::value_type           value_type;
    typedef std::valarray<value_type>                 vector_type;
    typedef FreeEnergyModel<model_type>               freeenergy_type;
    enum merit_enum {Tm, DeltaHS, TmDeltaHS, RTmDeltaHS};

    freeenergy_type Regression;
    model_type      Calculation;
    bool Debug;
    merit_enum Merit;

    GibbsMinimization(void): Regression(), Calculation(), Debug(false), Merit(Tm)
      {
      Regression.model=&Calculation;
      };
      
    inline  void merit_tm(void) {Merit=Tm;}
    inline  void merit_deltahs(void) {Merit=DeltaHS;}
    inline  void merit_tmdeltahs(void) {Merit=TmDeltaHS;}
    inline  void merit_rtmdeltahs(void) {Merit=RTmDeltaHS;}
    
    inline void print_merit_text(std::ostream& out)
      {
      switch (Merit)
        {
        case Tm       : out << "Merit function is (T-T')^2"; break;
        case DeltaHS  : out << "Merit function is (DeltaH-DeltaH')^2 + (DeltaS-DeltaS')^2"; break;
        case TmDeltaHS: out << "Merit function is (T-T')^2 + (DeltaH-DeltaH')^2 + (DeltaS-DeltaS')^2"; break;
        }
      out << std::endl;
      }

   inline std::string identification(void) {return std::string("Gibbs Free energy");}

   inline value_type evaluate(void)
     {
     Regression.predict_all();
     Regression.dataset.calculate_tm_difference_entire_set();
     if (Debug) std::cerr << Regression.dataset.Temperature.sqr_diff << " " << std::flush;
     value_type chi2;
     switch (Merit)
       {
       case Tm       : chi2 = Regression.dataset.Temperature.sqr_diff; break;
       case DeltaHS  : chi2 = Regression.dataset.Enthalpy.sqr_diff+Regression.dataset.Entropy.sqr_diff; break;
       case TmDeltaHS: chi2 = Regression.dataset.Temperature.sqr_diff+Regression.dataset.Enthalpy.sqr_diff+Regression.dataset.Entropy.sqr_diff; break;
       case RTmDeltaHS: chi2 = Regression.dataset.Temperature.relative_sqr_diff+Regression.dataset.Enthalpy.relative_sqr_diff+Regression.dataset.Entropy.relative_sqr_diff; break;
       }
     return chi2;
     }

   inline void print_verification(std::ostream& out)
     {
     Regression.verify(out);
     Regression.dataset.print_temperatures(out);
     print_merit_text(out);
     }

   inline void print_regression(std::ostream& out)
     {}

   inline void print_tex(std::ostream& out)
     {
     Regression.dataset.difference();
     Regression.print_tex(out);
     }

   inline void randomise_data(void)
     {
     Regression.dataset.randomise_dataset();
     }


   };
};
#endif
