// $Id: NucleotideDictionary.h 1119 2020-05-05 22:32:40Z ge $
// Copyright 2006 Gerald Weber
/// \file NucleotideDictionary.h
/// \brief Class which handles conversions of symbols to internal representations
///
/// $Revision: 1119 $
/// \author Gerald Weber <gweberbh@gmail.com>
#ifndef GBC_NUCLEOTIDEDICTIONARY_H
#define GBC_NUCLEOTIDEDICTIONARY_H "$Id: NucleotideDictionary.h 1119 2020-05-05 22:32:40Z ge $"
#include <fstream>
#include <map>
#include <list>
#include <complex>

namespace std {
template<class _Tp>
/// \brief Dummy operator to enable insertion of objects of type complex
/// \attention This is for compilation purposes only
bool operator<(const std::complex<_Tp>& cx1, const std::complex<_Tp>& cx2)
  {return true;}}

namespace gbc {
  
enum class NucleotideCode { empty, generic, DNA, RNA, IUPAC, Math };


/// \brief Conversion dictionary between symbols and internal representation of Nucleotide.
///
/// Implements a conversion dictionary for the symbols which are being read
/// from the data files to some internal representation _RepTp.
/// In a simplest case _RepTp could be char or int, but it can also be some
/// more elaborated structure.
/// _IndexTp has no duplicates, while _RepTp may have. For instance _RepTp could be both 'A' or 'a', which would
/// be converted to say 1. Example: add('A',1) and afterwards add('a',1). 
/// But 1 will always be converted back to the symbol that was defined first, in this case 'A'.
/// In summary, 'A' -> 1, 'a' -> 1, inverse: 1 -> 'A' but never 'a'.
  
template<class _RepTp=char, class _IndexTp=char>
class NucleotideDictionary
  {
  public:
    typedef _RepTp rep_type;
    typedef _IndexTp    index_type;
    typedef std::map<_IndexTp,_RepTp> index_map_type;
    typedef std::map<_RepTp,_IndexTp> rep_map_type;
    typedef std::map<_RepTp,_RepTp> complementary_map_type;
    typedef NucleotideDictionary<_RepTp,_IndexTp> DictionaryType;
    typedef std::list<_IndexTp> char_list_type;

    index_map_type index_map;
    rep_map_type rep_map; ///< Map for efficient inverse lookup
    complementary_map_type complementary_map; ///< Map for symbols that are complementary
    
    bool Debug; //Debug flag, set to true for seeing more information

    /// \brief Void constructor.
    NucleotideDictionary(void) : Debug(false) { populate(); };

    /// \brief Void constructor.
    NucleotideDictionary(NucleotideCode NC) : Debug(false) { populate(NC); };

//     /// \brief Void constructor.
//     NucleotideDictionary(bool t) : Debug(false) {};
    
    /// \brief Constructor which reads a dictionary file.
    ///
    /// The dictionary file is read immediately and only once.
    NucleotideDictionary(const char dict_name[]) : Debug(false)
      {
      read_dictionary(dict_name);
      };

   /// \brief Constructor which defines the dictionary conversions.
   ///
   /// This constructor gets rid of the need for a dictionary file.
   NucleotideDictionary(const _RepTp& N, ///< value for unknown (char='#')
                        const _RepTp& A, ///< value for adenine
			const _RepTp& C, ///< value for cytosine
			const _RepTp& G, ///< value for guanine
			const _RepTp& T) ///< value for thymine
                        : Debug(false)
     {
     add('#',N);
     add('A',A); add('C',C); add('G',G); add('T',T);
     add('a',A); add('c',C); add('g',G); add('t',T); //This guarantees that lowercase input is converted to uppercase
     set_complementary_rep(A,T);
     set_complementary_rep(C,G);
     }
     
   inline void populate(NucleotideCode NC=NucleotideCode::empty) {};

   /// \brief Adds new conversion rules to the dictionary.
   ///
   /// This method is simply a convenience replacement for std::map::insert.
   inline void add(const _IndexTp& smb,       //Indexing, usually char
                   const _RepTp& cnv,    //Representation
                   bool  replace_inverse=false //If true, inverse will be replace if it already exists
                   )  
      {
      if (Debug) std::cout << "smb=" << smb << " cnv=" << cnv << std::endl;
      if (index_map.find(smb) != index_map.end())
        std::cerr << std::endl << "Warning: replacing internal index " << smb << " with representation " << cnv << " " << __FILE__ << ":" << __LINE__ << std::endl;
      
      index_map[smb]=cnv;
      
      if (rep_map.find(cnv) != rep_map.end())
        {
        if (replace_inverse)
          {
          std::cerr << std::endl << "Warning: replacing representation " << cnv << " with internal index " << smb << " " << __FILE__ << ":" << __LINE__ << std::endl;
           rep_map[cnv]=smb;
          }
        }
      else rep_map[cnv]=smb;
      }
      
   inline void clear_all(void)
     {
     index_map.clear();
     rep_map.clear();
     complementary_map.clear();
     }

   /// \brief Adds new pair of complementary representations.
   // This assumes that the complementarity is symmetric, e.g., A is complementary to T, while T is complementary to A. 
   // However, this does not work for hybrids like DNA/RNA.
   // ATTENTION existing complementarities will be replaced
   inline void set_complementary_rep(const _RepTp& fst, ///< First internal symbol 
                                     const _RepTp& snd) ///< Complementary internal symbol
      {
      if ( (complementary_map.find(fst) != complementary_map.end()) or (complementary_map.find(snd) != complementary_map.end()) )
        std::cerr << std::endl << "Warning: replacing complementary for " << fst << ":" << snd << " " << __FILE__ << ":" << __LINE__ << std::endl;
      
      complementary_map[fst]=snd;
      complementary_map[snd]=fst;
      }
      
   inline void set_complementary_index(const _IndexTp& fst, ///< First internal symbol 
                                       const _IndexTp& snd) ///< Complementary internal symbol
      {
      if ( (complementary_map.find(index2rep(fst)) != complementary_map.end()) or (complementary_map.find(index2rep(snd)) != complementary_map.end()) )
        std::cerr << std::endl << "Warning: replacing complementary for " << fst << ":" << snd << " " << __FILE__ << ":" << __LINE__ << std::endl;
      
      complementary_map[index2rep(fst)]=index2rep(snd);
      complementary_map[index2rep(snd)]=index2rep(fst);
      }

   // Legacy function, use set_complementary_rep instead
   inline void complementary_pair(const _RepTp& fst, ///< First internal symbol 
                                  const _RepTp& snd) ///< Complementary internal symbol
      {
      set_complementary_rep(fst,snd);
      }


   /// \brief Conversion to complementary symbols found in complementary_map.
   ///
   /// \return Returns the complementary symbol of type char.
   inline char complementary_symbol(const char& fst) const ///< Symbol to convert
      {
      return rep2index(complementary(index2rep(fst)));
      }

   inline _RepTp complementary(const _RepTp& cnv,            ///< Symbol for which to find the complementary
                               bool  complain=true) const    ///< If true, raise exception when complementary not found
      {
      auto it=complementary_map.find(cnv);
      try
        {
        if (it != complementary_map.end()) return it->second;
        else if(complain) throw cnv;
        }
      catch (_RepTp errcnv)
        {
        std::cerr << std::endl << "Warning: complementary representation for \"" << errcnv <<  "\" not found " << __FILE__ << ":" << __LINE__ << std::endl;
        }
      return _RepTp();
      }

   /// \brief Removes conversion rules from the dictionary
   inline void rm(const char& smb)   ///< Symbol to be removed
      {
      _RepTp fst=index2rep(smb);
      _RepTp snd=rep2index(complementary(fst,false)); //complain=false as complementary may not exist
      if (complementary_map.find(fst) != complementary_map.end()) complementary_map.erase(complementary_map.find(fst));
      if (complementary_map.find(snd) != complementary_map.end()) complementary_map.erase(complementary_map.find(snd));
      rep_map.erase(index2rep(smb));
      index_map.erase(smb);
      }

    /// \brief Returns the _IndexTp of an _RepTp
    inline _IndexTp rep2index(const _RepTp& cnv) const ///< object of type _RepTp to be converted
      {
      if (not rep_map.empty()) //Not all types of NucleotideDictionary do have an inverse map
        {
        auto found=rep_map.find(cnv);
        if (found != rep_map.end()) return rep_map.find(cnv)->second;
        else if (Debug) std::cerr << std::endl << "Warning: no inverse symbol found for \"" << cnv << "\" " << __FILE__ << ":" << __LINE__ << std::endl;
        }
      else
        {
        for(auto dt=index_map.begin(); dt != index_map.end(); dt++)
          {
	  if (dt->second==cnv) return dt->first;
	  }
        }
      return _IndexTp();
      }

    // Legacy function, calls rep2index
    inline _IndexTp invert(const _RepTp& cnv) const ///< object of type _RepTp to be converted
      {
      return rep2index(cnv);
      }
      
    /// \brief Translate _IndexT to the _RepTp
    ///
    /// This looks up the translation map, if not found returns '#'.
    inline _RepTp index2rep(const _IndexTp& cr) const ///< char to be translated
      {
      auto fnd=index_map.find(cr);
      if (fnd != index_map.end()) return fnd->second; 
        else 
          {
          fnd=index_map.find('#');
          if (fnd != index_map.end()) return fnd->second; 
          }
      return _RepTp();
      }

    /// \brief Legacy function, calls index2rep
    inline _RepTp translate(const _IndexTp& cr) const ///< char to be translated
      {
      return index2rep(cr);
      }

    /// \brief Extractor which loads a whole dictionary from file.
    ///
    /// Use like: file >> dict;
    friend std::istream& operator>>(std::istream &fl, DictionaryType &dict)
      {
      while (!fl.eof() && fl.good())
        {
        _IndexTp symbol; _RepTp conversion;
        fl >> symbol >> conversion;
        dict.add(symbol,conversion);
        }
      return fl;
      }

    /// \brief Writes the contents of the dictionary to ostream.
    friend std::ostream& operator<<(std::ostream& os, const DictionaryType &dict)
      {
      for(auto dit : dict.index_map)
        os << dit.first << " " << dit.second << std::endl;
      return os;
      }

    /// \brief Reads a dictionary file
    void read_dictionary(const char dict_name[]) ///< name of the dictionary file
      {
      std::ifstream dictionary_file;
      dictionary_file.exceptions(std::ios_base::failbit | std::ios_base::badbit);
      try {dictionary_file.open(dict_name);}
      catch (const std::ios_base::failure& error)      {std::cerr << error.what() << __FILE__ << __LINE__ << std::endl;}
      catch (...)                       {std::cerr << "Unknown exception" << __FILE__ << __LINE__ << std::endl;}
      dictionary_file.exceptions(std::ios_base::goodbit);
      dictionary_file >> (*this);
      dictionary_file.close();
      }

    /// \brief Lists all characters known to be valid.
    char_list_type valid_characters(void)
      {
      char_list_type char_list;
      for (auto it  = index_map.begin();  it != index_map.end(); it++)
        {
        if (it->first != '#')
          {
          char_list.push_back(rep2index(it->second));
	  }
        }
      char_list.sort();
      char_list.unique();
      return char_list;
      }

  };
  
  //Specializations
   template<>
   inline void NucleotideDictionary<int,char>::populate(NucleotideCode NC)
     {
     switch (NC)
       {
       case NucleotideCode::empty   : break;
       case NucleotideCode::generic :
       case NucleotideCode::DNA     :
         add('#',0);
         add('A',1); add('C',2); add('G',3); add('T',4);
         add('a',1); add('c',2); add('g',3); add('t',4); //This guarantees that lowercase input is converted to uppercase
         set_complementary_index('A','T');
         set_complementary_index('C','G');
         break;
         
       case NucleotideCode::RNA :
         add('#',0);
         add('A',1); add('C',2); add('G',3); add('U',4);
         add('a',1); add('c',2); add('g',3); add('u',4); //This guarantees that lowercase input is converted to uppercase
         set_complementary_index('A','U');
         set_complementary_index('C','G');
         break;
         
       case NucleotideCode::IUPAC :
         add('A',1);    add('a',rep2index('A'));
         add('C',10);   add('c',rep2index('C'));
         add('G',100);  add('g',rep2index('G'));
         add('T',1000); add('t',rep2index('T'));
         add('R',index2rep('A') + index2rep('G')); add('r',rep2index('R')); //A or G
         add('Y',index2rep('C') + index2rep('T')); add('y',rep2index('Y')); //C or T
         add('S',index2rep('G') + index2rep('C')); add('s',rep2index('S')); //G or C
         add('W',index2rep('A') + index2rep('T')); add('w',rep2index('W')); //A or T
         add('K',index2rep('G') + index2rep('T')); add('k',rep2index('K')); //G or T
         add('M',index2rep('A') + index2rep('C')); add('m',rep2index('M')); //A or C
         add('B',index2rep('C') + index2rep('G') + index2rep('T')); add('b',rep2index('B')); //C or G or T
         add('D',index2rep('A') + index2rep('G') + index2rep('T')); add('d',rep2index('D')); //A or G or T
         add('H',index2rep('A') + index2rep('C') + index2rep('T')); add('h',rep2index('H')); //A or C or T
         add('V',index2rep('A') + index2rep('C') + index2rep('G')); add('v',rep2index('V')); //A or C or G
         add('N',index2rep('A') + index2rep('C') + index2rep('G') + index2rep('T'));  add('n',rep2index('N')); //any base
  
         set_complementary_index('A','T');
         set_complementary_index('C','G');
         set_complementary_index('R','Y');
         set_complementary_index('W','S');
         set_complementary_index('K','M');
         set_complementary_index('B','V');
         set_complementary_index('D','H');
         set_complementary_index('N','N');
         break;
         
       case NucleotideCode::Math :
         std::cerr << "Error: no NucleotideCode::Math implementation for NucleotideDictionary<int,char> " << __FILE__ << ":" << __LINE__ << std::endl;
         exit(1);
         
       default: break;
       }
     }
  
   template<>
   inline void NucleotideDictionary<char,char>::populate(NucleotideCode NC)
     {
     switch (NC)
       {
       case NucleotideCode::empty   : break;
       case NucleotideCode::generic :
       case NucleotideCode::DNA :
         add('#','N');
         add('A','A'); add('C','C'); add('G','G'); add('T','T');
         add('a','A'); add('c','C'); add('g','G'); add('t','T'); //This guarantees that lowercase input is converted to uppercase
         set_complementary_index('A','T');
         set_complementary_index('C','G');
         break;

       case NucleotideCode::RNA :
         add('#','N');
         add('A','A'); add('C','C'); add('G','G'); add('U','U');
         add('a','A'); add('c','C'); add('g','G'); add('u','U'); //This guarantees that lowercase input is converted to uppercase
         set_complementary_index('A','U');
         set_complementary_index('C','G');
         break;
         
       //https://www.bioinformatics.org/sms/iupac.html
       case NucleotideCode::IUPAC :
         add('A','A'); add('a','A');
         add('C','C'); add('c','C');
         add('G','G'); add('g','G');
         add('T','T'); add('t','T');
         add('R','R'); add('r','R'); //A or G
         add('Y','Y'); add('y','Y'); //C or T
         add('S','S'); add('s','S'); //G or C
         add('W','W'); add('w','W'); //A or T
         add('K','K'); add('k','K'); //G or T
         add('M','M'); add('m','M'); //A or C
         add('B','B'); add('b','B'); //C or G or T
         add('D','D'); add('d','D'); //A or G or T
         add('H','H'); add('h','H'); //A or C or T
         add('V','V'); add('v','V'); //A or C or G
         add('N','N'); add('n','N'); //any base
  
         set_complementary_index('A','T');
         set_complementary_index('C','G');
         set_complementary_index('R','Y');
         set_complementary_index('W','S');
         set_complementary_index('K','M');
         set_complementary_index('B','V');
         set_complementary_index('D','H');
         set_complementary_index('N','N');
         break;
         
       case NucleotideCode::Math :
         std::cerr << "Error: no NucleotideCode::Math implementation for NucleotideDictionary<char,char> " << __FILE__ << ":" << __LINE__ << std::endl;
         exit(1);
         
       default: break;
       }
     }
     
   template<>
   /// \brief Specialization of type complex<int>.
   inline void NucleotideDictionary<std::complex<int>,char>::add(const char& smb,const std::complex<int>& cnv, bool  replace_inverse)
     {
     index_map[smb]=cnv;
     }

   template<>
   /// \brief Specialization of type complex<double>.
   inline void NucleotideDictionary<std::complex<double>,char>::add(const char& smb,const std::complex<double>& cnv, bool  replace_inverse)
     {
     index_map[smb]=cnv;
     }

     template<>
   /// \brief Specialization of type complex<int>.
   inline std::complex<int> NucleotideDictionary<std::complex<int>,char>::complementary(const std::complex<int>& cnv, bool  complain) const ///< Symbol to convert
      {
      return conj(cnv);
      }
      

}
#endif
