// $Id: ParameterSearch.h 1197 2021-08-10 20:28:16Z ge $
/// \file ParameterSearch.h
/// \brief contains the ParameterSearch function class
///
/// $Revision: 1197 $
/// \author Gerald Weber <gweberbh@gmail.com>
#ifndef GBC_PARAMETERSEARCH_H
#define GBC_PARAMETERSEARCH_H "$Id: ParameterSearch.h 1197 2021-08-10 20:28:16Z ge $"
#include "Minimization.h"
#include "GSLMinimization.h"
#include <cmath>
#include <boost/regex.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/iostreams/filtering_streambuf.hpp>
#include <boost/iostreams/filter/gzip.hpp>
#include <boost/iostreams/device/back_inserter.hpp>
#include <boost/iostreams/copy.hpp>
#include "CPUtime.h"


namespace gbc
{
template<class _value_type>
class VariableLimits
  {
  public:
    typedef _value_type value_type;
    typedef std::deque<std::string> associated_deque_type;
    value_type current;
    value_type start;
    value_type stop;
    bool limits_set;
    associated_deque_type Associated_variables;//Variables that should have the same value as the main one

  VariableLimits(void): current(), start(), stop(), limits_set(false) {};
  
  void set_limits(value_type st, value_type en)
    {
    start=st;
    stop=en;
    limits_set=true;
    }
    
  value_type new_random_value(const double &rnd, const bool &linear)
    {
    double dice;
    if (limits_set)
      {
      dice=linear_ran();
      current = (stop-start)*dice + start;
      }
    else
      {
      if (linear) dice=linear_ran(2*rnd,0.5); //this will return a random value between -rnd and +rnd
      else        dice=gauss_ran(rnd);
      current *=(1.0+dice); //for linear this results in 1-rnd and 1+rnd
      }
    return  current;
    }

    
  };
  
  
template<class _EvaluateTp>
class ParameterSearch
  {
  public:
    typedef _EvaluateTp evaluate_type;
    typedef typename evaluate_type::value_type value_type;
    typedef std::valarray<value_type> vector_type;
    typedef std::map<std::string,VariableLimits<value_type> >     map_type;
    typedef std::map<size_t,map_type>        all_map_type;
    class files_class
      {
      public:
      std::ofstream Result;
      std::ofstream Parameter;
      std::ofstream Verification;
      std::ofstream Variation;
      std::ostream Log;//for gz compression
      std::ofstream Log_of;//for gz compression
      std::ofstream Regression;
      std::ofstream Evolution;
      std::string BaseName, VersionBase;
      int Version;
      bool Make_proceed_file;
      bool Make_files;
      bool Versioned;
      bool Log_compress;//for gz compression
      boost::iostreams::filtering_streambuf< boost::iostreams::output> log_filter;//for gz compression https://www.boost.org/doc/libs/1_46_1/libs/iostreams/doc/classes/filtering_streambuf.html
      
      files_class(void) : Version(0), Make_proceed_file(true), Make_files(true), Versioned(false), Log_compress(false), Log(Log_of.rdbuf()) {};
      
      inline void next_version(void)
        {
        close_all();
        Version++;
        std::stringstream ver;
        ver << "." << std::setw(5) << std::setfill('0') << Version;
        VersionBase=ver.str();
        open_all_version();
        }
      inline void overwrite(void)
        {
        close_all();
        if (Versioned) open_all_version(); else open_all();
        }
      inline void open(std::string bsname)
        {
        BaseName=bsname;
        if (Versioned) open_all_version(); else open_all();
        }
        
      //for gz compression we need to open Log_of in binary and append the gzip_compressor to the log_filter buffer
      inline void log_open(std::string name)
        {
        if (Log_compress)
          {
          log_filter.reset();
          Log.rdbuf(&log_filter);
          Log_of.open((name+".gz").c_str(),std::ofstream::binary);
          log_filter.push(boost::iostreams::gzip_compressor());//needs to be pushed before file
          log_filter.push(Log_of);
          }
        else
          {
          Log_of.open(name.c_str());
          }
        }
        
      inline void open_all(void)
        {
        if (not Make_files) return;
        Result.open((BaseName+".dat").c_str());
        Parameter.open((BaseName+".par").c_str());
        Verification.open((BaseName+".ver").c_str());
        Variation.open((BaseName+".var").c_str());
        log_open((BaseName+".log").c_str());
        Regression.open((BaseName+".reg").c_str());
        Evolution.open((BaseName+".average_std_evolution").c_str());

        std::ofstream ProcFile;
        if (Make_proceed_file) 
          {
          ProcFile.open((BaseName+".proceed").c_str());
          ProcFile << "1" << std::endl;
          ProcFile.close();
          }
        }
      inline void open_all_version(void)
        {
        if (not Make_files) return;
        Result.open((BaseName+".dat"+VersionBase).c_str());
        Parameter.open((BaseName+".par"+VersionBase).c_str());
        Verification.open((BaseName+".ver"+VersionBase).c_str());
        Variation.open((BaseName+".var"+VersionBase).c_str());
        log_open((BaseName+".log"+VersionBase).c_str());
        Regression.open((BaseName+".reg"+VersionBase).c_str());
        }

      inline void close_all(void)
        {
        if (not Make_files) return;//if we didn't open we do not close either
        Result.close();
        Parameter.close();
        Verification.close();
        Variation.close();
        //Log_of.close(); Do not call close or the program never stops.https://stackoverflow.com/questions/17844534/param-syntax-for-substituting-boost-filtering-stream-for-stdofstream
        boost::iostreams::close(log_filter);Log_of.close();
        Regression.close();
        Evolution.close();
        }
      inline bool proceed(void)
        {
        std::ifstream Proceed;
        Proceed.open((BaseName+".proceed").c_str());
        bool proc;
        Proceed >> proc;
        Proceed.close();
        return proc;
        }
      };
    

    enum minimization_engine_enum {GBC,GSL};
    minimization_engine_enum MinimizationEngine;
    /// The minimization class takes as template a class which performs the merit function evaluation
    Minimization<ParameterSearch<evaluate_type> > Minimize;
    GSLMinimization<ParameterSearch<evaluate_type> > GSLMinimize;

    evaluate_type Evaluation; ///< This is the function that will run in minimization

    bool Debug;
    bool Random_search;
    bool Proceed;
    bool Avoid_negative_parameters;
    bool Randomise_initial_parameters;
    bool Initial_parameters_linear; //if true randomization is linear, if false gaussian
    bool Randomise_dataset;
    bool Generate_files_for_each_round;
    value_type Initial_parameter_variation;
    int MaxEvaluations;
    int MaxMinimizations;
    int MaxRandomMinimizations; ///< Maximum number of random regression minimization, 0 means go on forever
    long int Evaluations;
    value_type Total_average_difference;
    value_type Total_standard_deviation_difference;
    value_type Limit_average_difference;
    value_type Limit_standard_deviation_difference;
    map_type Variable_map;
    map_type Best_variable_map;
    map_type Average_variable_map;
    map_type Previous_average_variable_map;
    map_type Standard_deviation_variable_map;
    map_type Previous_standard_deviation_variable_map;
    all_map_type Complete_variable_map;
    vector_type Best_vector;
    value_type  Best_value;
    std::string BaseFileName;
    files_class File, PermanentFile, BestValueFile;
    Range<value_type> Parameter_range;

    ParameterSearch(void)
      : MinimizationEngine(GBC), Minimize(), GSLMinimize(), Evaluation(), Debug(false), Random_search(false), Proceed(true), Avoid_negative_parameters(true),
        Randomise_initial_parameters(false), Initial_parameters_linear(true), Randomise_dataset(true), Generate_files_for_each_round(true), Initial_parameter_variation(value_type()),
        MaxEvaluations(1000), MaxMinimizations(1), MaxRandomMinimizations(0), Evaluations(), 
        Total_average_difference(), Total_standard_deviation_difference(), Limit_average_difference(), Limit_standard_deviation_difference()
         {}
        
    void minimization_method(std::string option)
      {
      if (option == std::string("gbc")) MinimizationEngine=GBC;
      else 
        {
        if (option == std::string("gsl")) MinimizationEngine=GSL;
        else 
          std::cerr << __FILE__ << ":" << __LINE__ << "Unknown minimization engine=" << option << ", options are: gbc or gsl" << std::endl;
        }
      }
      
    // Inserts a new value p in Variable_map and in the calculation map
    inline void update_calculation_parameter_map(const std::string id, const value_type& p)
      {
      Evaluation.Calculation.parameter_map.set_parameters(id,p);//updates the calculation map
      Variable_map[id].current=p; //updates the variables map
      for(size_t i=0; i < Variable_map[id].Associated_variables.size(); i++)
        Evaluation.Calculation.parameter_map.set_parameters(Variable_map[id].Associated_variables[i],p);//updates all associated variables
      }
      

    /// \brief Assing list of parameters which will vary during the minimization
    ///
    /// This will assign the current definition of the parameter to its 
    /// corresponding position in Variable_map.
    /// The initial value of Variable_map is whatever was stored in ParameterMap initially.
    inline void assign_variables(const std::deque<std::string>& var) ///< List of parameters, usually given on command line
      {
      const boost::regex regex_limits("^(.+)\\(([+-]?\\d+\\.?\\d*[eE]?[-+]?\\d*)\\|([+-]?\\d+\\.?\\d*[eE]?[-+]?\\d*)\\)");// Experimental/untested reserved for future implementation
      const boost::regex incomplete("^(.+)\\((.*)\\)");
      boost::smatch found;
      std::string variable;
      typename std::deque<std::string>::const_iterator var_list_it;
      for(var_list_it=var.begin(); var_list_it!= var.end(); ++var_list_it)
        {
        variable=*var_list_it;
        if (variable == std::string("T")) 
          Variable_map[variable].current=Evaluation.Calculation.temperature(); // If parameter is simply a T, assing temperature to it.
        else 
          {
          if (boost::regex_search(variable,found,regex_limits)) // ATTENTION Experimental/untested reserved for future implementation
            {
            std::string v=found[1], l1=found[2], l2=found[3];
            Variable_map[v].set_limits(std::atof(l1.c_str()),std::atof(l2.c_str()));
            std::cout << "\nAssigning interval to " << v << " (" << l1 << "|" << l2 << ")" << std::endl; 
            Evaluation.Calculation.parameter_map.get_parameter(v,Variable_map[v].current);
            }
          else
            {
            if (boost::regex_search(variable,found,incomplete)) // ATTENTION Experimental/untested reserved for future implementation
              {
              std::cerr << __FILE__ << " " << __LINE__ << "Error: could not parse " <<  variable << std::endl;
              }
            else
              {
              //Variables may be concatenated by equal sign as in AT_AT:entropy=AT_TA:entropy=AT_CG:entropy=AT_GC:entropy
              //The following will split these, the first variable is the lead parameter, all others become
              //associated variables which are placed into the Associated_variables deque. Whenever the lead parameter changes
              //all associated variables take its value.
              std::vector<std::string> strs;
              boost::split(strs,variable,boost::is_any_of("="));
            
              variable=strs[0];//assign the lead parameter (the first before any =)
              Evaluation.Calculation.parameter_map.get_parameter(variable,Variable_map[variable].current);//assigns the lead parameter
              
              std::cout << "Variable parameter = " << variable;
              if (strs.size() > 1) std::cout << " associated variables are: ";
              for(size_t i=1; i < strs.size(); i++) 
                {
                Variable_map[variable].Associated_variables.push_back(strs[i]);
                std::cout << strs[i] << " ";
                }
              std::cout << std::endl;
              }
            }
          }
        if (Debug) 
          {
          std::cout << __FILE__ << "(assign_variables)" << __LINE__ << " "
                    << variable << " " << Variable_map[variable].current;
          for(size_t i=0; i < Variable_map[variable].Associated_variables.size(); i++) std::cout << " = " << Variable_map[variable].Associated_variables[i];
          std::cout << "\n" << std::endl;
          }
          
        }
      } 

    /// \brief Replace values in ParameterMap with current values of Variable_map
    /// ATTENTION **** Apparently not in use
    inline void variable_map_reassign(void)
      {
      typename map_type::iterator it;
      for(it=Variable_map.begin(); it!= Variable_map.end(); ++it)
        {
        if (it->first == std::string("T")) 
          it->second.current=Evaluation.Calculation.temperature(); // If parameter is simply a T, assign temperature to it.
        else 
          update_calculation_parameter_map(it->first,it->second.current);
        if (Debug) std::cout << __FILE__ << "(variable_map_reassign)" << __LINE__  
                             << it->first << " "<< Variable_map[it->first].current << "\n" << std::endl;
        }
      } 

    /// \brief Randomize the initial values of VariableMap from ParameterMap.
    inline void variable_map_randomise(double rnd)
      {
      typename map_type::iterator it;
      for(it=Variable_map.begin(); it!= Variable_map.end(); ++it)
        {
        if (it->first == std::string("T")) 
          it->second.current=Evaluation.Calculation.temperature(); // If parameter is simply a T, assing temperature to it.
        else
	  {
          it->second.new_random_value(rnd,Initial_parameters_linear);
          if (Debug) std::cout << __FILE__ << "(variable_map_randomise)" << __LINE__ << "\n" 
                               << it->first << "<--->" << it->second.current << "\n" << std::endl;
	  }
        if (Debug) std::cout << __FILE__ << "(variable_map_randomise)" << __LINE__ << "\n" 
                             << it->first << " "<< Variable_map[it->first].current << "\n" << std::endl;
        }
      } 

    inline void basefile(std::string bsname)
      {
      BaseFileName=bsname;
      }

    inline void log_compress(const bool cps)
      {
      File.Log_compress=PermanentFile.Log_compress=BestValueFile.Log_compress=cps;
      }
      
    /// \brief Assigns the vector from the Variable_map
    inline void matrix2vector(vector_type& vec)
      {
      if (vec.size() != Variable_map.size()) vec.resize(Variable_map.size());
      typename map_type::iterator it;
      size_t n=0;
      for(it=Variable_map.begin(), n=0; it != Variable_map.end(); ++it, ++n)
        vec[n]=it->second.current;
      }

    /// \brief Assigns the contents of a vector to the Variable_map
    ///
    /// The vector is assigned in the same order as the variables are stored.
    inline void vector2var(const vector_type& vec)
      {
      typename map_type::iterator it;
      size_t n=0;
      for(it=Variable_map.begin(), n=0; it != Variable_map.end(); ++it, ++n)
        {
        it->second.current=vec[n];
        }
      }

    inline vector_type randomise(value_type scale,const vector_type& vec)
      {
      vector_type nvec(vec.size());
      size_t n=0;
      for(n=0; n < vec.size(); ++n)
        {
        double dice=rand()/static_cast<double>(RAND_MAX)-0.5;
        nvec[n]=vec[n]+dice*scale*vec[n];
        }
      return nvec;
      }

    inline vector_type random_search(value_type scale, const vector_type& vec)
      {
      vector_type nvec=randomise(scale,vec);
      value_type nv=evaluate(nvec);
      int evaluations=1;
      while ((nv > Best_value) && (evaluations < MaxEvaluations))
        {
        nvec=randomise(scale,vec);
        nv=evaluate(nvec);
        evaluations++;
        }
      if (nv < Best_value) 
        {
        File.Log << "random_search: " << evaluations << " ";
        print_res(File.Log,nv);
        return nvec;
        }
      else return vec;
      }


    /// \brief Evaluates the function from the vector or parameters
    ///
    /// First we assign the new parameters to the parameter map of the HeterogenousTM class.
    /// At the same time we update the Variable_map.
    /// Then the HeterogenousTM issues a get_parameters() call which load the modified
    /// parameters into the ModelPotentials. Only then we do the Evaluation which
    /// is defined in the SearchAction class.
    /// \Attention This returns Evaluate*negative_parameter_penalty
    inline value_type evaluate(const vector_type& vec)
     {
     value_type negative_parameter_penalty=1.0;
     typename map_type::iterator var_map_it;
     size_t n=0;
     for(var_map_it=Variable_map.begin(), n=0; var_map_it != Variable_map.end(); ++var_map_it, ++n)
       {
       if (var_map_it->first == std::string("T"))
         {
         Evaluation.Calculation.temperature(vec[n]);
         var_map_it->second.current=vec[n];
         }
       else
         {
         update_calculation_parameter_map(var_map_it->first,vec[n]);
         }
       
       if (Avoid_negative_parameters && (vec[n] < 0.0)) negative_parameter_penalty += 1.0;//this is a hack really
       }
     //if (Debug) std::cout << Calculation << std::endl;
     if (Debug) 
       {
       std::cout << __FILE__ << " " << __LINE__ << "\n"; 
       Evaluation.Calculation.parameter_map.print_parameters_and_stats(std::cout); 
       std::cout << std::endl;
       }
     Evaluation.Calculation.set_TM_parameters();
     Evaluations++;
     value_type evalresult=Evaluation.evaluate();
     if (Debug) std::cout << __FILE__ << " " << __LINE__ << " eval=" 
                          << evalresult << std::endl;
     return negative_parameter_penalty*evalresult;
     }

    /// \brief Evaluates the function from the Variable_map
    ///
    /// The Variable_map is saved into the ParameterMap of HeterogenousTM.
    /// Then the HeterogenousTM issues a set_TM_parameters() call which load the modified
    /// parameters into the ModelPotentials. Only then we do the Evaluation which
    /// is defined in the SearchAction class.
    inline value_type evaluate(void)
     {
     typename map_type::const_iterator var_map_it;
     for(var_map_it=Variable_map.begin(); var_map_it != Variable_map.end(); ++var_map_it)
       {
       update_calculation_parameter_map(var_map_it->first,var_map_it->second.current);
       }
     if (Debug) 
       {
       std::cout << __FILE__ << " " << __LINE__ << "\n"; 
       Evaluation.Calculation.parameter_map.print_parameters_and_stats(std::cout); 
       std::cout << std::endl;
       }
     Evaluation.Calculation.set_TM_parameters();
     return Evaluation.evaluate();
     }

    /// \brief Scans the parameters for the Variable_map 
    ///
    /// We increment the parameters in Variable_map sequentially and run evaluate.
    /// \Attention This function can be very time consuming if the ranges have
    /// a large number of steps or if there are too many variables in Variable_map.
    inline void scan_parameters(Range<value_type>& prange)
      {
      value_type Best_sqrdiff=1e+12, sqrdiff;

      File.open(BaseFileName);
      typename std::vector<Range<value_type> > ranges;
      prange.reset();
      typename map_type::iterator it;
      for(it=Variable_map.begin(); it != Variable_map.end(); ++it)
        {
        ranges.push_back(prange);
        it->second.current=prange.begin();
        File.Log << it->first << " " << prange << " " << prange.steps() << " ";
        }
      File.Log << std::endl;

      for(prange.reset(); !prange.finished(); ++prange) File.Log << prange.current() << " ";
      File.Log << std::endl;

      it=Variable_map.begin(); ++it; File.Log << it->second.current << " ";
      sqrdiff=evaluate();
      File.Log << sqrdiff << " ";
      if (sqrdiff < Best_sqrdiff)
        {
        Best_sqrdiff=sqrdiff;
        Best_variable_map=Variable_map;
        }

      typename std::vector<Range<value_type> >::iterator next=ranges.begin();
      ++next;
      typename std::vector<Range<value_type> >::iterator loop=ranges.begin();
      bool finished=false;
      typename map_type::iterator variable;
      variable=Variable_map.begin();
      ++variable;
      while (!finished)
        {
        if (loop->finished())
          {
          File.Log << std::endl;
          loop->reset(); Variable_map.begin()->second.current=loop->current();
          if (next != ranges.end())
            {
            if (next->finished())
              {
              ++next; ++variable; if (next == ranges.end()) break;
              }
            else
              {
              next->next_step(); variable->second.current=next->current();
              File.Log << variable->second.current << " ";
              }
            }
          else break;
          }
        else
          {
          loop->next_step(); Variable_map.begin()->second.current=loop->current();
          }

        if (!loop->finished() && !next->finished())
          {
          sqrdiff=evaluate();
          File.Log << sqrdiff << " ";
          if (sqrdiff < Best_sqrdiff)
            {
            Best_sqrdiff=sqrdiff;
            Best_variable_map=Variable_map;
            }
          }

        }

      Variable_map=Best_variable_map;
      evaluate();

      print_var(File.Variation,"");
      print_par(File.Parameter,"");
      print_best_value(File.Verification);
      Evaluation.print_verification(File.Verification);

      }

    /// \brief Scans the parameters for the Variable_map 
    ///
    /// We increment the parameters in Variable_map sequentially and run evaluate.
    /// \Attention This function can be very time consuming if the ranges have
    /// a large number of steps or if there are too many variables in Variable_map.
    inline void scan_parameters(Range<value_type>& prange1, Range<value_type>& prange2)
      {
      value_type Best_sqrdiff=1e+12, sqrdiff;

      File.open(BaseFileName);
      typename std::vector<Range<value_type> > ranges;
      prange1.reset();
      prange2.reset();
      typename map_type::iterator it=Variable_map.begin(); 
      ranges.push_back(prange1);
      it->second.current=prange1.begin();
      File.Log << it->first << " " << prange1 << " " << prange1.steps() << " ";
      ++it;
      ranges.push_back(prange2);
      it->second.current=prange2.begin();
      File.Log << it->first << " " << prange2 << " " << prange2.steps() << " ";
      File.Log << std::endl;

      for(prange1.reset(); !prange1.finished(); ++prange1) File.Log << prange1.current() << " ";
      File.Log << std::endl;

      it=Variable_map.begin(); ++it; File.Log << it->second.current << " ";
      sqrdiff=evaluate();
      File.Log << sqrdiff << " ";
      if (sqrdiff < Best_sqrdiff)
        {
        Best_sqrdiff=sqrdiff;
        Best_variable_map=Variable_map;
        }

      typename std::vector<Range<value_type> >::iterator next=ranges.begin();
      ++next;
      typename std::vector<Range<value_type> >::iterator loop=ranges.begin();
      bool finished=false;
      typename map_type::iterator variable;
      variable=Variable_map.begin();
      ++variable;
      while (!finished)
        {
        if (loop->finished())
          {
          File.Log << std::endl;
          loop->reset(); Variable_map.begin()->second.current=loop->current();
          if (next != ranges.end())
            {
            if (next->finished())
              {
              ++next; ++variable; if (next == ranges.end()) break;
              }
            else
              {
              next->next_step(); variable->second.current=next->current();
              File.Log << variable->second.current << " ";
              }
            }
          else break;
          }
        else
          {
          loop->next_step(); Variable_map.begin()->second.current=loop->current();
          }

        if (!loop->finished() && !next->finished())
          {
          sqrdiff=evaluate();
          File.Log << sqrdiff << " ";
          if (sqrdiff < Best_sqrdiff)
            {
            Best_sqrdiff=sqrdiff;
            Best_variable_map=Variable_map;
            }
          }

        }

      Variable_map=Best_variable_map;
      evaluate();

      print_var(File.Variation,"");
      print_par(File.Parameter,"");
      print_best_value(File.Verification);
      Evaluation.print_verification(File.Verification);

      }
      
    inline void set_print_log_no_improvement(bool sp)
      {
      Minimize.PrintLogNoImprovement = GSLMinimize.PrintLogNoImprovement = sp;
      }

    inline void set_print_log_progress(bool sp)
      {
      Minimize.PrintLogProgress = GSLMinimize.PrintLogProgress = sp;
      }
      
    inline void set_log_progress_interval(int secs)
      {
      Minimize.ProgressInterval = GSLMinimize.ProgressInterval = secs;
      }

    inline void set_print_log_replace_worst(bool sp)
      {
      Minimize.PrintLogReplaceWorst = sp;
      }

    inline void minimize(Range<double>& clength,const double& limit)
      {
      //Evaluation.Debug=true;
      value_type cl=clength.current();//relates to option -cl
      vector_type  p; 
      matrix2vector(p);//here we place the numerical content of the variable map (option -var) into a vector
      Best_vector.resize(p.size());
      Best_value=std::numeric_limits<value_type>::max();
      value_type previous=Best_value, diff=std::numeric_limits<value_type>::max();
      int improove=0;
      evaluate(p);
      //MaxMinimization relates to option -mm 
      //limit relates to option -limit
      while((diff > limit) && (improove < MaxMinimizations))
        {
      File.Log << "*** Starting round " << improove << std::endl;
      switch (MinimizationEngine)
          {
          case GBC: Minimize.reset(*this,p,cl);
                    Minimize.MaxEvaluations=MaxEvaluations;
                    Minimize.run(limit);
                    if (Best_value > Minimize.best_value())
                      {            
                      Best_value=Minimize.best_value();
                      Best_vector=Minimize.best();
                      }
                    break;
          case GSL: GSLMinimize.reset(*this,p,cl);
                    GSLMinimize.MaxEvaluations=MaxEvaluations;
                    GSLMinimize.run(limit);
                    if (Best_value > GSLMinimize.best_value())
                      {            
                      Best_value=GSLMinimize.best_value();
                      Best_vector=GSLMinimize.best();
                      }
                    break;
          };
        diff=fabs(Best_value-previous);
        previous=Best_value;
        improove++;
        if (Random_search) 
          {
          p=random_search(0.5,Best_vector);
          File.Log << "*** best vector randomized to 0.5" << std::endl;
          }
        else p=Best_vector;
        }
      if (diff <= limit) File.Log << "*** Difference between the last two rounds " << diff << " below " << limit << ", stopping minimization altogether" << std::endl; 
      File.Log << "*** Local minimization stopped at round " << improove << " (maximum rounds was "<< MaxMinimizations << ")" << std::endl;
      evaluate(Best_vector);
      vector2var(Best_vector);
      }

   inline void minimization(std::string par_id, double salt, Range<double>& clength,const double& limit)
     {
     File.open(BaseFileName);
     print_par(File.Verification,par_id);
     File.Verification << "Calculation:" << std::endl;
     Evaluation.Calculation.print_calculation(File.Verification);
     Evaluation.Calculation.parameter_map.print_parameters_and_stats(File.Verification,"Calculation.parameter_map:"); 
     File.Verification << std::endl;
     Minimize.LogFile=&File.Log;
     GSLMinimize.LogFile=&File.Log;
     Evaluation.print_verification(File.Verification);
     minimize(clength,limit);
     print_res(File.Result,salt);
     print_var(File.Variation,par_id);
     print_par(File.Parameter,par_id);
     print_best_value(File.Verification);
     Evaluation.print_verification(File.Verification);
     Evaluation.print_regression(File.Regression);
     File.close_all();
     }

   /// \brief Performs a minimization on randomized datasets
   inline void minimization_random(int random_seed,        ///< Randomization seed
                                   std::string par_id,     ///< parameter identification
                                   double salt,            ///< salt concentration, only used for printing out results, has otherwise no effect
                                   Range<double>& clength, ///< minimization parameter characteristic length
                                   const double& limit)    ///< minimization parameter limit below which we stop searching for better values
     {
     ParameterMap<value_type> pm=Evaluation.Calculation.parameter_map; //Here we store the initial parameters
     map_type vm=Variable_map; //Here we store the initial Variable Map
     srand(random_seed);
     PermanentFile.open(BaseFileName);
     PermanentFile.Log << "Random seed=" << random_seed << std::endl;
     File.BaseName=BaseFileName;
     File.Versioned=true;
     File.Make_files=Generate_files_for_each_round;
     BestValueFile.BaseName=BaseFileName;
     BestValueFile.Versioned=true;
     BestValueFile.VersionBase=".best";
     Proceed=PermanentFile.proceed();
     Evaluation.print_verification(PermanentFile.Verification);
     PermanentFile.Verification 
       << Evaluation.Regression.dataset.Temperature.error << " "
       << Evaluation.Regression.dataset.Enthalpy.error << " "
       << Evaluation.Regression.dataset.Entropy.error << std::endl;
     value_type Global_best_value=1e+10;
     while(Proceed)
       {
       File.next_version();
       if (Randomise_initial_parameters)
         {
         //Evaluation.Calculation.parameter_map.randomise(Initial_parameter_variation);
         //variable_map_reassign();
         variable_map_randomise(Initial_parameter_variation);
         std::cout << "Initial variables randomised for minimization round " << File.Version << std::endl;
         print_cpu();//so that we can have an immediate access to how much memory was used
         }
       print_par(File.Verification,par_id);
       File.Verification << "Calculation:" << std::endl;
       Evaluation.Calculation.print_calculation(File.Verification);
       Evaluation.Calculation.parameter_map.print_parameters_and_stats(File.Verification,"Calculation.parameter_map:"); 
       File.Verification << std::endl;
       Minimize.LogFile=&File.Log;
       GSLMinimize.LogFile=&File.Log;
       if (Randomise_dataset)
         {
         Evaluation.Regression.dataset.randomise_dataset();
         File.Verification << Evaluation.Regression.dataset.Temperature.std_deviation << " "
                           << Evaluation.Regression.dataset.Enthalpy.std_deviation << " "
                           << Evaluation.Regression.dataset.Entropy.std_deviation << std::endl;
         PermanentFile.Verification << Evaluation.Regression.dataset.Temperature.std_deviation << " "
                           << Evaluation.Regression.dataset.Enthalpy.std_deviation << " "
                           << Evaluation.Regression.dataset.Entropy.std_deviation << std::endl;
	 }
       Evaluation.print_verification(File.Verification);
       minimize(clength,limit); //here is the part that takes time to run
       
       Complete_variable_map[File.Version]=Variable_map;

       print_res(File.Result,salt);
       print_res(PermanentFile.Result,salt);
       print_var(File.Variation,par_id);
       print_par(File.Parameter,par_id);
       print_best_value(File.Verification);
       Evaluation.print_verification(File.Verification);
       Evaluation.print_regression(File.Regression);
       
       if (Global_best_value > Best_value)
         {
         std::cout << "New best set of parameters at round " << File.Version << " with chi2 = " << Best_value << std::endl;
         Global_best_value = Best_value;
         BestValueFile.overwrite();
         print_res(BestValueFile.Result,salt);
         print_var(BestValueFile.Variation,par_id);
         print_par(BestValueFile.Parameter,par_id);
         print_best_value(BestValueFile.Verification);
         Evaluation.print_verification(BestValueFile.Verification);
         Evaluation.print_regression(BestValueFile.Regression);
         BestValueFile.close_all();
         }
       
       //From here we reset everything to start over
       Evaluation.Calculation.parameter_map=pm; // restore initial parameters
       Evaluation.Calculation.set_TM_parameters();
       Variable_map=vm; //restore initial variable map
       Proceed=PermanentFile.proceed();
       this->save_av_par_file(par_id,File.Version);
       
       if (MaxRandomMinimizations and (File.Version >= MaxRandomMinimizations)) 
         {
         std::cout << "Maximum number of randomized local minimizations reached (max=" << File.Version << ")" << std::endl;
         break;
         }
       if ((Total_average_difference <  Limit_average_difference) and (File.Version >= 10)) 
         {
         std::cout << "Limit for total average difference reached " << Total_average_difference
         << " limit= " << Limit_average_difference << ", stoped at round " << File.Version << std::endl;
         break;
         }
      if ((Total_standard_deviation_difference <  Limit_standard_deviation_difference) and (File.Version >= 10)) 
         {
         std::cout << "Limit for total standard deviation difference reached " << Total_standard_deviation_difference
         << " limit= " << Limit_standard_deviation_difference << ", stoped at round " << File.Version << std::endl;
         break;
         }
         
       }
     }
     
   inline void make_averages(void)
     {
     typename all_map_type::iterator mp;
     typename map_type::iterator var;
     Average_variable_map.clear();
     Standard_deviation_variable_map.clear();
     for(mp = Complete_variable_map.begin(); mp != Complete_variable_map.end(); ++mp)
       {
       for(var=mp->second.begin(); var != mp->second.end(); ++var)
         {
         Average_variable_map[var->first].current += var->second.current;
         }
       }
    for(var=Average_variable_map.begin(); var != Average_variable_map.end(); ++var)
      {
      var->second.current /= Complete_variable_map.size();
      }
    for(mp = Complete_variable_map.begin(); mp != Complete_variable_map.end(); ++mp)
       {
       for(var=mp->second.begin(); var != mp->second.end(); ++var)
         {
         Standard_deviation_variable_map[var->first].current += pow(Average_variable_map[var->first].current-var->second.current,2);
         }
       }
    for(var=Standard_deviation_variable_map.begin(); var != Standard_deviation_variable_map.end(); ++var)
      {
      var->second.current = sqrt(var->second.current/Complete_variable_map.size());  
      }
    }

   //Prints out the variable map and its associated variables too
   inline void print_var(std::ostream& out, std::string ident)
      {
      out << ident << std::endl;
      typename map_type::iterator it;
      for(it=Variable_map.begin(); it != Variable_map.end(); ++it)
        {
        out << it->first << " " << it->second.current << std::endl;
        for(size_t i=0; i < it->second.Associated_variables.size(); i++) out << it->second.Associated_variables[i] << " " << it->second.current << std::endl;
        }
      }

    inline void print_par(std::ostream& out, std::string ident)
      {
      Evaluation.Calculation.parameter_map.print_parameters_and_stats(out,ident);
      out << std::endl;
      }

    inline void save_av_par_file(std::string indent, int version)
      {
      make_averages();
      std::ofstream out;
      out.open((BaseFileName+".par.average").c_str());
      out << indent << "-average-" << version << std::endl;
      typename map_type::iterator var;
      for(var=Average_variable_map.begin(); var != Average_variable_map.end(); ++var)
        {
        out << var->first << " " << var->second.current << " +/- " << Standard_deviation_variable_map[var->first].current;
        if (Previous_average_variable_map[var->first].current != 0)
          {
          value_type average_difference=fabs((Previous_average_variable_map[var->first].current-var->second.current)/Previous_average_variable_map[var->first].current);
          out << " [" << average_difference;
          Total_average_difference += average_difference;
          }
        if (Previous_standard_deviation_variable_map[var->first].current != 0)
          {
          value_type std_difference=fabs((Previous_standard_deviation_variable_map[var->first].current-Standard_deviation_variable_map[var->first].current)/Previous_standard_deviation_variable_map[var->first].current);
          out << " " << std_difference << "]";
          Total_standard_deviation_difference += std_difference;
          }
        out << std::endl;
        
        }
      out.close();
      
      Previous_average_variable_map=Average_variable_map;
      Previous_standard_deviation_variable_map=Standard_deviation_variable_map;
     
      Total_average_difference /= Average_variable_map.size();
      Total_standard_deviation_difference /= Average_variable_map.size();
      
      PermanentFile.Evolution << Complete_variable_map.size() << " " << Total_average_difference << " " << Total_standard_deviation_difference << std::endl;
      }

    inline void print_head(std::ostream& out)
      {
      typename map_type::iterator it;
      out << Best_value << "Best ";
      for(it=Variable_map.begin(); it != Variable_map.end(); ++it)
        out << it->first << " ";
      out << std::endl;
      }


    inline void print_res(std::ostream& out, value_type sl)
      {
      typename map_type::iterator it;
      if (sl != value_type()) out << sl << " ";
      out << Best_value << " ";
      for(it=Variable_map.begin(); it != Variable_map.end(); ++it)
        out << it->second.current << " ";
      out << std::endl;
      }

    inline void print_best_value(std::ostream& out)
      {
      out << Best_value << std::endl;
      }

  };
}
#endif
