// $Id: vargibbs.cpp 1201 2021-08-13 20:15:05Z ge $
/// \file var-gibbs.cpp
/// \brief program for searching parameters of Gibbs Free Energy Models
/// \author Gerald Weber <gweberbh@gmail.com>
/// $Revision: 1201 $

#include <valarray>
#include "Options.h"
#include "NeighbourSequence.h"
#include <unistd.h>
#include "FreeEnergyModel.h"
#include "ParameterSearch.h" 
#include "GibbsMinimization.h"

using namespace gbc;
using namespace std;

typedef double value_type;

typedef ParameterSearch<GibbsMinimization<IndividualNearestNeighbours<double> > > min_function_type;
 

// We need to declare an object of type min_function_type to avoid compiler errors
// of inclomplete type
min_function_type place_holder;
template<> min_function_type *GSLMinimization<min_function_type>::Function=NULL;

#if !defined(SVNVERSION)
#define SVNVERSION "$Revision: 1201 $"
#endif
#if !defined(LINUXSYSTEM)
#define LINUXSYSTEM "unknown"
#endif 
#if !defined(BUILDTIME)
#define BUILDTIME __DATE__
#endif 
#ifndef COMPILERNAME
#ifdef __INTEL_COMPILER
#define COMPILERNAME "icpc"    
#else
#define COMPILERNAME "gcc"
#endif
#endif 

int main(int argc,char *argv[])
  {
  std::cout << "*** vargibbs version=" << std::string(SVNVERSION) << " system=" << std::string(LINUXSYSTEM) << " builddate=" << std::string(BUILDTIME) << " compiler=" << std::string(COMPILERNAME) << std::endl;
  Nucleotide<>::dictionary.add('_','_');
  Nucleotide<>::dictionary.add('U','U');
 
  Option<string>         output_name("o","output files basename, resulting files will be appended with .dat, .ver etc",OptionMap::necessary);
  Option<deque<string> > par_file("par","list of parameter file",OptionMap::necessary);
  Option<deque<string> > datafile("data","list of sequence data files",OptionMap::optional);
  OptionMap::add_synonymn("xml","data");//For compatibility with var-par

  Option<string>         duplextype("duplextype","type of duplex, e.g. DNA or RNA",OptionMap::optional,"DNA");
  Option<double>         temp_err("ee","experimental temperature error in celsius",OptionMap::optional,0.0);
  Option<int>            random_seed("rs","random_seed",OptionMap::optional,1);

  Option<double>         salt("seqsalt","salt concentration (needed for salt correction of isolated sequences)",OptionMap::optional);
  Option<double>         org_tm("tm","original Tm (needed for salt correction of isolated sequences)",OptionMap::optional);
  Option<double>         target_salt("targetsalt","target salt concentration (for converting Tm)",OptionMap::optional);
  Option<string>         salt_scheme("saltscheme","type of salt conversion scheme",OptionMap::optional,"nocorrection");
  salt_scheme.add_restricted_option("owczarzy04eq19","owczarzy04eq19 for DNA");
  salt_scheme.add_restricted_option("owczarzy04eq20","owczarzy04eq20 for DNA");
  salt_scheme.add_restricted_option("owczarzy04eq21","owczarzy04eq21 for DNA");
  salt_scheme.add_restricted_option("owczarzy04eq22","owczarzy04eq22 for DNA");
  salt_scheme.add_restricted_option("chen13eq19","chen13eq19 for RNA");
  salt_scheme.add_restricted_option("chen13eq20","chen13eq20 for RNA");
  salt_scheme.add_restricted_option("chen13eq21","chen13eq21 for RNA");
  salt_scheme.add_restricted_option("chen13eq22","chen13eq22 for RNA");
  salt_scheme.add_restricted_option("schildkraut65","old correction for DNA");
  salt_scheme.add_restricted_option("nakano99","nakano99 for DNA/RNA only for 1000 to 100 Na+");

  Option<bool>           prediction_with_salt_correction("predsaltcorr","apply salt correction to predicted Tm (default: apply to dataset Tm)",OptionMap::optional,false);
  
  Option<deque<double> > species_concentration("ct","single or list of species concentration Ct (in 10^-6 M)",OptionMap::optional);
  Option<double>         multiply_sp_conc("mct","multiply species concentration Ct by a given factor",OptionMap::optional,1.0);
  Option<string>         calculation("calc","which type of calculation to perform",OptionMap::necessary,"prediction");
  calculation.add_restricted_option("nncheck","interprets the sequences in the -data file and exits without performing calculations");
  calculation.add_restricted_option("minlocal","runs local minimization");
  calculation.add_restricted_option("minglobal","runs global minimization, needs -mrr");
  calculation.add_restricted_option("prediction","performs simple Tm prediction");
  
  
  Option<deque<string> > variable("var","list of variables to loop",OptionMap::optional);
  Option<string>         minimization_method("minmethod","select select minimization method",OptionMap::optional,"gsl");
  minimization_method.add_restricted_option("gsl","uses GNU scientific library minimization function");
  minimization_method.add_restricted_option("gbc","uses own minimization algorithm");
  Option<string>         minimization_log_level("minloglevel","selects the amount of information for the minimization log", OptionMap::optional,"progress");
  minimization_log_level.add_restricted_option("progress","log minimization every time (see -loginterval) a better chi2 is achieved");
  minimization_log_level.add_restricted_option("full","log minimization for every evaluation, even if chi2 does not improve");
  minimization_log_level.add_restricted_option("terse","only log minimization for last chi of each round");
  Option<int>            log_level_interval("loginterval","in seconds, interval of -minloglevel=progress",0);
  
  Option<bool>           randomize_dataset("randomizedataset","randomize dataset",OptionMap::optional,false);
  Option<int>            max_random_rounds("mrr","maximum number of local minimization rounds (option calc=minglobal)",OptionMap::optional,10);
  Option<map<string,string> > dict("dict","Additions to nucleotide dictionary",OptionMap::optional);
  Option<double>         limit("limit","minimization limit",OptionMap::optional,0.01);
  Option<Range<double> > clength("cl","characteristic length",OptionMap::optional,10.0);
  Option<int>            max_eval("me","maximum number of evaluations",OptionMap::optional,std::numeric_limits<int>::max());
  Option<int>            max_min("mm","maximum number of minimizations",OptionMap::optional,2);
  Option<double>         av_diff_limit("adl","limit for total average difference",OptionMap::optional);
  Option<double>         av_stdev_limit("asl","limit for total standard deviation difference",OptionMap::optional);
  Option<bool>           generate_n_files("gn","generate files for each round of minimization",OptionMap::optional,true);
  Option<bool>           nn_parameters_only("nnpo","use NN parameters only, this ignores initialization, terminal and symmetry parameters",OptionMap::optional,false);
  Option<bool>           log_compress("zlog","compress all log files in gzip format",OptionMap::optional,false);

  Option<string>         par_id("parid","parameter id",OptionMap::optional,"vargibbs");
  Option<string>         debug("debug","select debuging",OptionMap::optional,"none");
  debug.add_restricted_option("none","no debuging");
  debug.add_restricted_option("parameter","Debugs parameter calculation");
  debug.add_restricted_option("minimization","Debugs minimization module");
  
  Option<string>         sequence("seq","sequence main strand 5'->3'",OptionMap::optional);
  Option<string>         csequence("cseq","sequence complementary strand 3'->5'",OptionMap::optional);
  
  Option<string>         merit("merit","merit function",OptionMap::optional,"Tm");
  merit.add_restricted_option("Tm","use Tm");
  merit.add_restricted_option("DeltaHS","use DeltaH and DeltaS");
  merit.add_restricted_option("TmDeltaHS","use Tm, DeltaH and DeltaS");
  
  Option<double>         random_initial_values("randominitial","random initial values",0.0);

  //CG_equivalent and AT_equivalent will be compared to BP_reduced_duplex in sequencestructure/SequenceInfo.h
  Option<deque<string> > CG_equivalent("cgeq","Additional base pairs equivalent to CG, used for calculating fCG",OptionMap::optional); 
  Option<deque<string> > AT_equivalent("ateq","Additional base pairs equivalent to AT, used for Is_AT_only",OptionMap::optional); 

  Option<bool>           printusedpar("printusedpar","print parameters that were used with file extension .usedpar",OptionMap::optional,false);

  //Unused option, reserved for future use
//   Option<int>            seq_length("sl","fixed sequence length",OptionMap::optional,0);

  
  OptionMap::arguments(argc,argv);
  OptionMap::scan_arguments();
  if ((calculation.check("minlocal")) || (calculation.check("minglobal")))
    {
    variable.turn_necessary(true);
    if (calculation.check("minglobal"))
      {
      temp_err.turn_necessary(randomize_dataset.Value==1);
      random_initial_values.turn_necessary(!randomize_dataset.Provided);
      randomize_dataset.turn_necessary(!random_initial_values.Provided);
      }
    }
  if (calculation.check("prediction"))
    {
    species_concentration.turn_necessary(sequence.Provided);
    datafile.turn_optional(sequence.Provided);
    sequence.turn_necessary(!datafile.Provided);
    }
  if (calculation.check("nncheck"))
    {
    output_name.turn_optional(true);
    par_file.turn_optional(true);
    if (not sequence.provided()) datafile.turn_necessary(true); else datafile.turn_necessary(false);
    }
  if (target_salt.provided() or prediction_with_salt_correction.Value == true)
    {
    salt_scheme.turn_necessary(true);
    if (sequence.provided()) salt.turn_necessary(true);
    }
  
  
  OptionMap::check_if_all_given();
  OptionMap::show_options();

  if (dict.provided())
    {
    for(auto mit : dict.Value)
      {
      std::cout << "Adding new nucleotide to dictionary= "  << mit.first[0] << " and its complementary= " << mit.second[0] << std::endl;
      Nucleotide<>::dictionary.add(mit.first[0],mit.first[0],true);
      Nucleotide<>::dictionary.add(mit.second[0],mit.second[0],true);
      Nucleotide<>::dictionary.complementary_pair(mit.second[0],mit.first[0]);
      }
    }

  if (duplextype.check("RNA"))
    {
    Nucleotide<>::dictionary.complementary_pair('A','U');
    std::cout << Nucleotide<>::dictionary;
    }


  min_function_type pregress;
  if (minimization_method.provided()) pregress.minimization_method(minimization_method.Value);
  pregress.Avoid_negative_parameters=false;
  
  if (merit.provided())
    {
    if (merit.check("Tm")) pregress.Evaluation.merit_tm();
    if (merit.check("DeltaHS")) pregress.Evaluation.merit_deltahs();
    if (merit.check("TmDeltaHS")) pregress.Evaluation.merit_tmdeltahs();
    if (merit.check("RTmDeltaHS")) pregress.Evaluation.merit_rtmdeltahs();
    pregress.Evaluation.print_merit_text(std::cout);
    }

  pregress.MaxEvaluations=max_eval;
  
  pregress.MaxMinimizations=max_min;

  pregress.Evaluation.Calculation.NN_parameters_only=nn_parameters_only.Value;
    
  if (max_random_rounds.provided()) pregress.MaxRandomMinimizations=max_random_rounds.Value;
  
  if (av_diff_limit.provided()) pregress.Limit_average_difference=av_diff_limit.Value;
  if (av_stdev_limit.provided()) pregress.Limit_standard_deviation_difference=av_stdev_limit.Value;
  
  if (minimization_log_level.check("full"))
    {
    pregress.set_print_log_no_improvement(true);
    pregress.set_print_log_progress(true);
    pregress.set_print_log_replace_worst(true);
    }
    
  if (minimization_log_level.check("progress"))
    {
    pregress.set_print_log_no_improvement(false);
    pregress.set_print_log_progress(true);
    pregress.set_print_log_replace_worst(false);
    }
    
  if (minimization_log_level.check("terse"))
    {
    pregress.set_print_log_no_improvement(false);
    pregress.set_print_log_progress(false);
    pregress.set_print_log_replace_worst(false);
    }
    
  if (log_level_interval.provided()) pregress.set_log_progress_interval(log_level_interval.Value);
    
  if (debug.provided()) 
    {
    if (debug.check("parameter")) pregress.Evaluation.Calculation.Debug=true;
    if (debug.check("minimization")) pregress.Debug=true;
    }
    


  if(calculation.check("prediction") || calculation.check("nncheck"))
    {
    IndividualNearestNeighbours<> gfe;
    FreeEnergyModel<IndividualNearestNeighbours<> > gibbs;
    gibbs.model=&gfe;
    gfe.NN_parameters_only=nn_parameters_only.Value;
    std::string outfilefilename=output_name.Value+".dat";
    ofstream outfile; outfile.open(outfilefilename.c_str());
    ofstream verfile; verfile.open(((string)output_name+".ver").c_str());
    std::string nncheckfilename=output_name.Value+".nncheck";
    if (debug.provided()) 
      {
      if (debug.check("parameter")) gfe.Debug=true;
      }

    ofstream nncheckfile; 
    if (calculation.check("nncheck")) nncheckfile.open(nncheckfilename.c_str());
    
    if (CG_equivalent.provided())
      for (auto cg : CG_equivalent.Value)  gibbs.dataset.CG_equivalent_set.insert(cg);
    if (AT_equivalent.provided())
      for (auto at : AT_equivalent.Value)  gibbs.dataset.AT_equivalent_set.insert(at);

    gibbs.read_rules_and_parameters_from_list_of_files(par_file.Value);

    if (salt_scheme.provided() and datafile.provided()) 
      {
      gibbs.dataset.select_salt_correction(salt_scheme);
      gibbs.dataset.Target_salt_ct=target_salt.Value;
      if (gibbs.dataset.Target_salt_ct == value_type())
        std::cout << "Info: -targetsalt=0, using salt concentration in dataset as target for salt correction" << std::endl; 
      gibbs.dataset.Prediction_with_salt_correction=prediction_with_salt_correction.Value;
    
      std::cout << "Info: applying salt correction scheme " << salt_scheme.Value;
      if (gibbs.dataset.Prediction_with_salt_correction) 
        std::cout << " to predicted Tm" << std::endl;
      else 
        std::cout << " to dataset" << std::endl;
      }

    if(datafile.provided())
      {
      gibbs.dataset.read_list_of_data_files(datafile.Value,species_concentration.Value,multiply_sp_conc.Value);
      if(calculation.check("nncheck"))
        {
        if (output_name.provided())
          {
          std::cout << "Info: Writing nncheck analysis to file " << nncheckfilename << std::endl;
          gibbs.dataset.print_temperatures(nncheckfile,true);
          if (target_salt.provided() and salt_scheme.provided()) 
            {
            std::cout << "Info: Writing salt corrections to file " << outfilefilename << std::endl;
            gibbs.dataset.print_adjusted_dataset(outfile);
            }
          }
        else
          {
          gibbs.dataset.print_temperatures(std::cout,true);
          }
        return 0;
        }
      }
    else
      {
      std::string cseq;
      if (csequence.provided()) cseq=(std::string)csequence;
      else
        {
        std::cout << "Info: complementary sequence not provided, working out from complementary nucleotides " << std::endl;
        NucleotideSequence<> seq(sequence.Value.length()),comp(sequence.Value.length());
        seq << sequence.Value;
        comp = seq;
        comp.complementary();
        cseq = (string)comp;
        }
      cout << "Calculating for single sequence " << sequence.Value << "/" << cseq << std::endl;
      if (salt_scheme.provided())
        {
        gibbs.dataset.select_salt_correction(salt_scheme);
        gibbs.dataset.Target_salt_ct=target_salt.Value;
        gibbs.dataset.Prediction_with_salt_correction=prediction_with_salt_correction.Value;
    
        std::cout << "Info: applying salt correction scheme " << salt_scheme.Value << std::endl;
        }
      if (salt.provided()) std::cout << "Info: sequence salt concentration=" << salt.Value << " mM" << std::endl;
      gibbs.dataset.add_temperature(sequence.Value,cseq,salt.Value,org_tm.Value,species_concentration.Value[0]);
      if (calculation.check("nncheck")) 
        {
        gibbs.dataset.print_temperatures(nncheckfile,true);
        print_cpu();
        return 0;
        }
      }
       
    gibbs.predict_all();
    gibbs.dataset.print_temperatures(outfile);
    gibbs.dataset.calculate_and_print_verify(verfile);
          
    if (printusedpar.check(true))
      {
      ofstream used_par_file; 
      used_par_file.open(((string)output_name+".usedpar").c_str());      
      gfe.parameter_map.print_parameters_and_stats(used_par_file,"parameters");
      printusedpar.Value=false; //to avoid print again
      }
    }

  if(calculation.check("minsimple") || calculation.check("minlocal"))
    {
    pregress.Evaluation.Regression.read_rules_and_parameters_from_list_of_files(par_file.Value);
    
    if(species_concentration.provided()) pregress.Evaluation.Regression.dataset.Species_concentration=species_concentration;
    if(datafile.provided()) pregress.Evaluation.Regression.dataset.read_list_of_data_files(datafile.Value);
    pregress.basefile(output_name);
    pregress.log_compress(log_compress);
    pregress.assign_variables(variable.Value);
    pregress.Generate_files_for_each_round=generate_n_files.Value;
    pregress.minimization(par_id,salt,clength.Value,limit.Value);//salt only used for printing
    }

  if(calculation.check("minmult") || calculation.check("minglobal"))
    {
    pregress.Evaluation.Regression.read_rules_and_parameters_from_list_of_files(par_file.Value);
    
    if(species_concentration.provided()) pregress.Evaluation.Regression.dataset.Species_concentration=species_concentration;
    if(datafile.provided()) pregress.Evaluation.Regression.dataset.read_list_of_data_files(datafile.Value);
    pregress.basefile(output_name);
    pregress.log_compress(log_compress);
    pregress.assign_variables(variable.Value);
    pregress.Generate_files_for_each_round=generate_n_files.Value;
    if (randomize_dataset.Value)
      {
      pregress.Randomise_dataset=randomize_dataset.Value;
      pregress.Evaluation.Regression.dataset.Temperature.error=temp_err;
      }
    if (random_initial_values.provided())
      {
      pregress.Randomise_initial_parameters=true;
      pregress.Initial_parameter_variation=random_initial_values.Value;
      }


    pregress.minimization_random(random_seed,par_id,salt,clength.Value,limit.Value);//salt only used for printing
    }

    if (printusedpar.check(true))
      {
      ofstream used_par_file; 
      used_par_file.open(((string)output_name+".usedpar").c_str());      
      pregress.Evaluation.Calculation.parameter_map.print_parameters_and_stats(used_par_file,"parameters");
      }

    print_cpu();
  return 0;
}

